import pandas as pd
import pickle
import pathlib
import logging

logger = logging.getLogger(__name__)

from .data import RawData, DataReader, PymzmlDataReader
from .sample import Sample
from .feature import Feature, FeatureCollection, FeatureTable, MzmineFeatureTable, PeakonlyFeatureTable, XcmsFeatureTable
from .annotation import AnnotationTable

class Experiment():
    """
    Simple representation of MS experiment.

    Contain the data and meta-data relative to an experiment, 
    no metadata specific to the experiment design is stored (such as biological groups).
    The location and list of the raw data files and feature file are stored within this class.
    The raw data (spectra, chromatogram and features) is stored in objects of type Sample and 
    FeatureTable directly accessible through their respective lists.

    """

    def __init__(self, raw_file_folder_path, feature_table_path, input_format, labels=['High_quality','Low_quality','Noise'], load_MS1=False, name="NeatMS_experiment"):
        self.name = name
        self.raw_file_folder = self.get_file_folder(raw_file_folder_path)
        self.raw_files = self.get_raw_files()
        self.feature_tables = []
        self.data_reader = self.set_data_reader(feature_table_path)
        self.samples = self.load_samples(load_MS1)
        self.input_format = input_format
        self.load_feature_table(feature_table_path, labels)


    def set_data_reader(self, feature_table_path):
        feature_table_file_folder = pathlib.Path(feature_table_path)
        # Test the feature table file format 
        if feature_table_file_folder.is_dir() or feature_table_file_folder.suffix.lower() == '.csv':
            # If the file format is .csv, use pymzml backend file reader
            # First check that raw files are indeed in .mzml format
            raw_file_type = set([file.suffix.lower() for file in self.raw_file_folder.iterdir()])
            # Temporary fix: sometimes an empty suffix is present (probably due to temporary files) we discard it for now
            # TODO: Create stable fix
            raw_file_type.discard('')
            if len(raw_file_type) > 1:
                logger.error('The raw file folder contains several file formats, only .mzml format is accepted')
                return None
            if list(raw_file_type)[0] != '.mzml':
                logger.error('Required format is .mzml. Provided format: %s',list(raw_file_type)[0])
                return None
            data_reader = PymzmlDataReader()
            logger.info('Data reader backend: %s', 'pymzml')
            return data_reader
        else:
            # Use the openms backend file reader
            # TODO: Feature table from openms needs to be define
            data_reader = OpenmsDataReader()
            logger.info('Data reader backend: %s', 'pyopnems')
            return data_reader


    def get_file_names(self):
        return [file.name for file in self.samples]
        
    
    def get_file_folder(self, file_folder_path):
        if file_folder_path != None:
            file_folder = pathlib.Path(file_folder_path)
            return file_folder
        return None
    

    def get_raw_files(self):
        if self.raw_file_folder != None:
            return [file for file in self.raw_file_folder.iterdir() if file.suffix.lower() in ['.mzxml','.mzml','.mzdata']]
        return None


    def load_samples(self, load_data=False):
        if self.raw_file_folder == None:
            logger.error('raw_file_folder_path: %s does not exist', self.raw_file_folder)
            return None
        if len(self.raw_files) == 0:
            logger.error('Raw folder is empty')
            return None
        file_count = 0
        samples = []
        for file in self.raw_files:
            file_count += 1
            logger.info('Loading file %d / %d', file_count, len(self.raw_files))
            sample = Sample(file, self.data_reader)
            if load_data:
                sample.get_MS1()
            samples.append(sample)
        self.samples = samples
        return samples


    # This contain code duplicate, need to be factorised
    def load_feature_table(self, feature_table_path, labels=['High_quality','Low_quality','Noise']):
        feature_table_file_folder = pathlib.Path(feature_table_path)
        # Test if the path is a folder (several files = unaligned features)
        if feature_table_file_folder.is_dir():
            # Test if provided files are all in .csv format 
            feature_table_file_type = set([file.suffix.lower() for file in feature_table_file_folder.iterdir()])
            if len(feature_table_file_type) > 1:
                logger.error('The feature table file folder contains several file formats, only .csv format is accepted')
                return None
            if list(feature_table_file_type)[0] != '.csv':
                logger.error('Required format is .csv. Provided format: %s',list(feature_table_file_type)[0])
                return None
            if self.input_format.lower() == 'mzmine':
                feature_table = MzmineFeatureTable(feature_table_path)
            elif self.input_format.lower() == 'peakonly':
                feature_table = PeakonlyFeatureTable(feature_table_path)
            elif self.input_format.lower() == 'xcms':
                logger.error('XCMS input should be a .csv file, directory given')
            self.feature_tables.append(feature_table)
            feature_table.load_feature_table()
            sample_names = self.get_file_names()
            feature_table.create_column_map(sample_names)
            feature_table.load_features(self.samples)
            annotation_table = AnnotationTable(feature_table, labels)
            feature_table.annotation_table = annotation_table
            return feature_table
        # Else test if feature table file format is csv (one file = aligned features) 
        elif feature_table_file_folder.suffix.lower() == '.csv':
            logger.info('Loading feature table: %s',feature_table_file_folder.resolve())
            # Default feature table loading
            if self.input_format.lower() == 'mzmine':
                feature_table = MzmineFeatureTable(feature_table_path)
            elif self.input_format.lower() == 'peakonly':
                logger.error('Aligned features from peakonly are not supported')
            elif self.input_format.lower() == 'xcms':
                feature_table = XcmsFeatureTable(feature_table_path)
            self.feature_tables.append(feature_table)
            feature_table.load_feature_table()
            sample_names = self.get_file_names()
            feature_table.create_column_map(sample_names)
            feature_table.load_features(self.samples)
            annotation_table = AnnotationTable(feature_table, labels)
            feature_table.annotation_table = annotation_table
            return feature_table
        else:
            # TODO: add support for openms .featurexml or .consensusfeaturexml file format 
            logger.error('Only .csv file format is currently supported for the feature table')

    def get_columns(self, export_properties):
        # Need to respect the export properties order (4 first colunms, the rest follow the input)
        columns = ['feature ID', 'sample']
        if "mz" in export_properties:
            columns.append('m/z')
        if "rt" in export_properties:
            columns.append('retention time')
        # Quick hack: TODO remove for loop
        for export_property in export_properties:
            if export_property == "peak_mz":
                columns.append("peak_mz")
            if export_property == "peak_rt":
                columns.append("peak_rt")
            if export_property == "peak_rt_start":
                columns.append("peak_rt_start")
            if export_property == "peak_rt_end":
                columns.append("peak_rt_end")
            if export_property == "peak_mz_min":
                columns.append("peak_mz_min")
            if export_property == "peak_mz_max":
                columns.append("peak_mz_max")
            if export_property == "label":
                columns.append("label")
            if export_property == "height":
                columns.append("height")
            if export_property == "area":
                columns.append("area")
            if export_property == "area_bc":
                columns.append("area_bc")
            if export_property == "sn":
                columns.append("sn")
        return columns


    def export_to_dataframe(self, export_classes = ["High_quality", "Low_quality","Noise"], min_group_classes = ["High_quality", "Low_quality","Noise"],min_group_size = 0, exclude = [], use_annotation = False, export_properties = ["rt", "mz", "height","area","label"], long_format = True):
        total_sample_number = len(self.samples)
        # Set the minimum feature size (number of aligned peaks) for the feature to be kept
        if (0 < min_group_size <= 1):
            min_group_size = min_group_size * total_sample_number
        # Adjust the number to 1 if "min_group_size == 0", most conservative option
        # if min_group_size == 0:
        #     min_group_size = 1
        # If the exclusion list is not empty
        if len(exclude) > 0:
            sample_list = []
            # Iterate through all the samples
            for sample in self.samples:
                # If the sample object or the sample name is not in the exclusion list, keep the sample (add to the list)
                if (sample not in exclude) & (sample.name not in exclude):
                    sample_list.append(sample)
        else:
            sample_list = self.samples
        feature_collection_list = self.feature_tables[0].feature_collection_list
        export_feature_collection_list = []
        for feature_collection in feature_collection_list:
            # First filter to exclude features present in fewer samples than "min_group_size"
            if len(feature_collection.feature_list) >= min_group_size:
                # Get the number of peak predicted (annotated) with the target classes
                peak_count = feature_collection.get_monoisotopic_peak_number(prediction_class=min_group_classes, use_annotation=use_annotation)
                if (peak_count > 0) and (peak_count >= min_group_size):
                    # Add the feature collection to the export list
                    export_feature_collection_list.append(feature_collection)
        export_data = []
        for feature_collection in export_feature_collection_list:
            feature_collection_data = feature_collection.export_data(sample_list = sample_list, export_classes = export_classes, use_annotation = use_annotation, export_properties = export_properties, long_format = long_format)
            if long_format:
                export_data = export_data + feature_collection_data
            else:
                export_data.append(feature_collection_data)

        if long_format:
            columns = self.get_columns(export_properties)
            dataframe = pd.DataFrame(export_data, columns = columns) 
        else:
            dataframe = pd.DataFrame(export_data)

        return dataframe


    def export_csv(self, filename, index=True, na_rep='', export_classes = ["High_quality", "Low_quality", "Noise"], min_group_classes = ["High_quality", "Low_quality", "Noise"], min_group_size = 0, exclude = [], use_annotation = False, export_properties = ["rt", "mz", "height","area","label"]):
        dataframe = self.export_to_dataframe(export_classes, min_group_classes, min_group_size, exclude, use_annotation, export_properties)
        dataframe.to_csv(filename, index=index, na_rep=na_rep)
        file_path = pathlib.Path(filename)
        logger.info('Exporting data to %s',file_path.resolve())


    def save(self):
        file_name = self.name + '.pkl'
        with open(file_name, 'wb') as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)




from .data import RawData

class Sample():

    """
    Representation of a MS sample contain in an experiment.

    Contains, peak and feature lists, raw_data object to store raw file and connect to appropriate backend reader. 
    """

    # Sample counter to automatically assign sample ids
    sample_counter = 0
    sample_dict = dict()

    # This class method returns the peak instance from a peak id
    # This is to avoid iteration through all instances of peaks to find a specific one
    @classmethod
    def get_sample(cls, name):
        return cls.sample_dict[name]
    
    def __init__(self, raw_file_path, data_reader):
        self.id = Sample.sample_counter
        self.raw_file = raw_file_path
        self.name = self.raw_file.stem
        self.file_name = self.raw_file.name
        self.peak_list = []
        self.feature_list = []
        self.raw_data = RawData(self, raw_file_path, data_reader)
        Sample.sample_counter += 1
        Sample.sample_dict[self.name] = self


    def get_MS1(self):
        return self.raw_data.get_MS1()


    def extract_chromatogram(self, rt_start, rt_end, mz_min, mz_max):
        return self.raw_data.extract_chromatogram(rt_start, rt_end, mz_min, mz_max)


    def create_interpolated_chromatograms(self, vals=120, margin=1, normalise=True, peak_list=None, min_scan_num=5):
        # By default, if no peak list is provided, all peaks are used 
        if peak_list is None:
            peak_list = self.peak_list

        self.raw_data.create_interpolated_chromatograms(vals, margin, normalise, peak_list, min_scan_num)





import pandas as pd
import numpy as np
import pathlib
import logging

logger = logging.getLogger(__name__)

from .peak import Peak

class Feature():

    """
    Representation of an MS feature.

    Can contains several peaks that are grouped together from the same sample (adduct, isotopic peak, fragment...),
    mz and RT attributes correspond to the monoisotopic peak.
    Attached to a feature collection object that contain features from other samples aligned with this instance.  
    """

    # Feature counter to automatically assign feature ids
    feature_counter = 0
    feature_dict = dict()

    # This class method returns the feature instance from a feature id
    # This is to avoid iteration through all instances of features to find a specific one
    @classmethod
    def get_feature(cls, secondary_id):
        return cls.feature_dict[secondary_id]

    def __init__(self, mz, RT, sample=None, secondary_id=None, feature_collection=None):
        self.id = Feature.feature_counter
        self.secondary_id = secondary_id
        self.mz = mz
        self.RT = RT
        self.sample = sample
        self.peak_list = []
        self.feature_collection = feature_collection
        Feature.feature_counter += 1
        if secondary_id != None:
            Feature.feature_dict[self.secondary_id] = self


class FeatureCollection():

    """
    Representation of features from several samples aligned together

    Only contain a list of feature.
    """

    # Feature collection counter to automatically assign feature collection ids
    feature_collection_counter = 0

    def __init__(self):
         self.id = FeatureCollection.feature_collection_counter
         self.feature_list = []
         FeatureCollection.feature_collection_counter += 1


    def get_monoisotopic_peak_number(self, prediction_class, use_annotation):
        peak_count = 0
        for feature in self.feature_list:
            for peak in feature.peak_list:
                if peak.valid:
                    if (peak.monoisotopic is None) | (peak.monoisotopic is True):
                        if use_annotation:
                            if peak.annotation:
                                if peak.annotation.label in prediction_class:
                                    peak_count += 1
                            else:
                                if peak.prediction.label in prediction_class:
                                    peak_count += 1
                        else:
                            if peak.prediction.label in prediction_class:
                                peak_count += 1
        return peak_count


    def export_data(self, sample_list, export_classes, use_annotation, export_properties, long_format):
        # ["rt", "mz", "peak_mz", "peak_rt", "peak_rt_start", "peak_rt_end", "peak_mz_min", "peak_mz_max", "label", "height", "area", "area_bc", "sn"]
        if long_format:
            feature_data_list = []
            feature_id = self.id
            for feature in self.feature_list:
                sample_str = feature.sample.name
                for peak in feature.peak_list:
                    if peak.valid:
                        if use_annotation and peak.annotation:
                            label = peak.annotation.label
                        else:
                            label = peak.prediction.label
                        # If label in export_classes we start adding info to list
                        if label in export_classes:
                            data_list = []
                            data_list.append(feature_id)
                            data_list.append(sample_str)
                            if "mz" in export_properties:
                                data_list.append(feature.mz)
                            if "rt" in export_properties:
                                data_list.append(feature.RT)
                            for export_property in export_properties:
                                if export_property == "peak_mz":
                                    data_list.append(peak.mz)
                                if export_property == "peak_rt":
                                    data_list.append(peak.RT)
                                if export_property == "peak_rt_start":
                                    data_list.append(peak.rt_start)
                                if export_property == "peak_rt_end":
                                    data_list.append(peak.rt_end)
                                if export_property == "peak_mz_min":
                                    data_list.append(peak.mz_min)
                                if export_property == "peak_mz_max":
                                    data_list.append(peak.mz_max)
                                if export_property == "label":
                                    data_list.append(label)
                                if export_property == "height":
                                    data_list.append(peak.height)
                                if export_property == "area":
                                    data_list.append(peak.area)
                                if export_property == "area_bc":
                                    data_list.append(peak.area_baseline_corrected)
                                if export_property == "sn":
                                    data_list.append(peak.sn)
                            feature_data_list.append(data_list)
            return feature_data_list
        else:
            data_dictionnary = dict()
        if "mz" in export_properties:
            data_dictionnary["m/z"] = self.feature_list[0].mz
        if "rt" in export_properties:
            data_dictionnary["retention time"] = self.feature_list[0].RT
        for feature in self.feature_list:
            if feature.sample in sample_list:
                sample_str = feature.sample.name
                for peak in feature.peak_list:
                    if peak.valid:
                        if use_annotation and peak.annotation:
                            label = peak.annotation.label
                        else:
                            label = peak.prediction.label
                        if label in export_classes:
                            if (peak.monoisotopic is None) | (peak.monoisotopic is True):
                                for export_property in export_properties:
                                    if export_property == "peak_mz":
                                        entry_str = ' '.join([sample_str,'Peak m/z'])
                                        data_dictionnary[entry_str] = peak.mz
                                    if export_property == "peak_rt":
                                        entry_str = ' '.join([sample_str,'Peak RT'])
                                        data_dictionnary[entry_str] = peak.RT
                                    if export_property == "peak_rt_start":
                                        entry_str = ' '.join([sample_str,'Peak RT start'])
                                        data_dictionnary[entry_str] = peak.rt_start
                                    if export_property == "peak_rt_end":
                                        entry_str = ' '.join([sample_str,'Peak RT end'])
                                        data_dictionnary[entry_str] = peak.rt_end
                                    if export_property == "peak_mz_min":
                                        entry_str = ' '.join([sample_str,'Peak m/z min'])
                                        data_dictionnary[entry_str] = peak.mz_min
                                    if export_property == "peak_mz_max":
                                        entry_str = ' '.join([sample_str,'Peak m/z max'])
                                        data_dictionnary[entry_str] = peak.mz_max
                                    if export_property == "label":
                                        entry_str = ' '.join([sample_str,'Peak label'])
                                        data_dictionnary[entry_str] = label
                                    if export_property == "height":
                                        entry_str = ' '.join([sample_str,'Peak height'])
                                        data_dictionnary[entry_str] = peak.height
                                    if export_property == "area":
                                        entry_str = ' '.join([sample_str,'Peak area'])
                                        data_dictionnary[entry_str] = peak.area
        return data_dictionnary


class FeatureTable():

    """
    Representation of the feature extraction performed on the dataset

    Contains the feature table input file and enable its loading.
    Automate the creation of features and features object.

    Every feature table input format requires a specific class deirved 
    from this general FeatureTable class with dedicated loading functions.
    """

    # Feature table counter to automatically assign feature table ids
    feature_table_counter = 0

    def __init__(self, feature_table_path=None, origin=None):
        self.id = FeatureTable.feature_table_counter
        self.feature_table_path = feature_table_path
        self.feature_table = None
        self.column_map = None
        self.annotation_table = None
        self.feature_collection_list = []
        self.origin = origin
        FeatureTable.feature_table_counter += 1



class MzmineFeatureTable(FeatureTable):

    """
    Inherits from FeatureTable.

    Mzmine specific feature table object.
    Enable reading mzmine aligned and unaligned feature tables. 
    """

    def __init__(self, feature_table_path=None, origin=None):
        logger.info('Feature table format: MZmine')
        super().__init__(feature_table_path, origin)


    def load_feature_table(self):
        logger.info('Loading feature table and converting format')
        feature_table_file_folder = pathlib.Path(self.feature_table_path)
        if self.feature_table_path == None:
            logger.error('Feature table path missing')
            return None
        # If the path is a directory (Unaligned peaks, one table per sample)
        elif feature_table_file_folder.is_dir():
            # List all .csv files
            feature_table_files = [file for file in feature_table_file_folder.iterdir() if file.suffix.lower() in ['.csv']]
            # Create an empty dataframe
            feature_table = pd.DataFrame()
            # Iterate through all files (samples)
            for feature_table_file in feature_table_files:
                # Read the sample specific feature table
                sample_feature_table = pd.read_csv(feature_table_file)
                # Join the table to the main dataframe
                feature_table = feature_table.append(sample_feature_table, sort=False)
        # Else simply read the file
        else:
            feature_table = pd.read_csv(self.feature_table_path)
        # Reindex the dataframe
        feature_table = feature_table.reset_index(drop=True)
        # Drop extra empty columns sometimes created by mzmine
        feature_table = feature_table.dropna(how='all', axis=1)
        # Replace all 0 values with NA
        feature_table = feature_table.replace(0, np.nan)
        # Add feature index column (Important: perform after NA replacement)
        feature_table['Feature_ID'] = feature_table.index
        # Create list with the new column names so we can change format (wide to long)
        new_column_names = ['row m/z','row retention time']
        test_list = ['end', 'max', 'min', 'start'] 
        for column_name in feature_table.columns[2:-1]:
            res = any(ele in column_name for ele in test_list)
            split_column_name = column_name.split(' ')
            split_column_name.append(split_column_name.pop(0))
            if res:
                ele = split_column_name[2]
                split_column_name.remove(ele)
                split_column_name.insert(1,ele)
            new_column_name = str(' ').join(split_column_name)
            new_column_names.append(new_column_name)
        new_column_names.append('Feature_ID')
        # Change column names
        feature_table.columns = new_column_names
        # Convert dataframe from wide to long
        feature_table = pd.wide_to_long(feature_table, ["Peak m/z", "Peak RT", "Peak start RT", "Peak end RT","Peak height","Peak area","Peak min m/z","Peak max m/z"], i="Feature_ID", j="sample", sep=" ", suffix='.+')
        # Remove all empty entries
        feature_table = feature_table.dropna(how='all', axis=0, subset=["Peak m/z", "Peak RT", "Peak start RT", "Peak end RT","Peak height","Peak area","Peak min m/z","Peak max m/z"])
        self.feature_table = feature_table
        return feature_table


    def create_column_map(self, samples):
        # Useless since we convert input to long format 
        return None


    def load_features(self, sample_list):
        sample_map = dict()
        for sample in sample_list:
            sample_map[sample.file_name] = sample

        peak_number = self.feature_table.shape[0]
        feature_number = self.feature_table.index.max()[0]
        i = 0

        logger.info('Loading %d features and %d peaks ', feature_number, peak_number)

        df_group = self.feature_table.groupby('Feature_ID')
        for name, group in df_group:
            feature_collection = FeatureCollection()
            for row_index, row in group.iterrows():
                sample = sample_map[row_index[1]]
                mz = row['row m/z']
                RT = row['row retention time']
                feature = Feature(mz, RT, sample, feature_collection=feature_collection)
                peak_dict = dict(
                    sample = sample,
                    RT = row['Peak RT'], 
                    mz = row['Peak m/z'],
                    rt_start = row['Peak start RT'],
                    rt_end = row['Peak end RT'],
                    mz_min = row['Peak min m/z'],
                    mz_max = row['Peak max m/z'],
                    height = row['Peak height'],
                    area = row['Peak area'],
                )
                # Create peak
                new_peak = Peak(sample, peak_dict['RT'], peak_dict['mz'], peak_dict['rt_start'], peak_dict['rt_end'], peak_dict['mz_min'], 
                    peak_dict['mz_max'], peak_dict['height'], peak_dict['area'])
                # Add feature to the peak (one peak is attached to a unique feature)
                new_peak.feature = feature
                # Add the peak to the sample
                sample.peak_list.append(new_peak)
                # Add the peak to the feature
                feature.peak_list.append(new_peak)
                # Add the feature to the sample
                sample.feature_list.append(feature)
                # Add the feature to the feature collection
                feature_collection.feature_list.append(feature)
            # Add the feature to the feature list (all peaks belonging to this feature have now been created)
            self.feature_collection_list.append(feature_collection)
        logger.info('Feature table loaded with success')
        return None


class PeakonlyFeatureTable(FeatureTable):

    """
    Inherits from FeatureTable.

    Peakonly specific feature table object.
    Enable reading peakonly unaligned feature tables.
    Aligned features not supported as peakonly does not support alignment 
    """

    def __init__(self, feature_table_path=None, origin=None):
        logger.info('Feature table format: Peakonly')
        super().__init__(feature_table_path, origin)


    def load_feature_table(self):
        feature_table_file_folder = pathlib.Path(self.feature_table_path)
        if self.feature_table_path == None:
            logger.error('Feature table path missing')
            return None
        # If the path is a directory (Unaligned peaks, one table per sample)
        elif feature_table_file_folder.is_dir():
            # List all .csv files
            feature_table_files = [file for file in feature_table_file_folder.iterdir() if file.suffix.lower() in ['.csv']]
            # Create an empty dataframe
            feature_table = pd.DataFrame()
            # Iterate through all files (samples)
            for feature_table_file in feature_table_files:
                # Read the sample specific feature table
                sample_feature_table = pd.read_csv(feature_table_file)
                # Reconstruct the name of the corresponding raw file
                raw_file_name = feature_table_file.stem + '.mzML'
                # Modify the table: Create an rt column
                sample_feature_table['rt'] = sample_feature_table[['rtmin', 'rtmax']].mean(axis=1)
                # Keep only the column required (solves formating issues when some Unnamed extra column is created)
                sample_feature_table = sample_feature_table[['mz','rt','mzmin','mzmax','rtmin', 'rtmax','intensity']]
                # Add sample name to all columns except mz and rt
                sample_feature_table.columns = ['mz','rt',raw_file_name+' Peak mzmin',raw_file_name+' Peak mzmax',raw_file_name+' Peak rtmin',raw_file_name+' Peak rtmax',raw_file_name+' Peak area']
                # Join the table to the main dataframe
                feature_table = feature_table.append(sample_feature_table, sort=False)
            self.feature_table = feature_table.fillna(0)
            return feature_table
        else:
            logger.error('Provide path to directory containing single sample feature tables')
            return None


    def create_column_map(self, samples):
        feature_samples = []
        for column in self.feature_table.columns[2:]:
            feature_samples.append(column.split('.')[0])

        column_mapping = dict()
        for sample in samples:
            column_mapping[sample] = dict()

        for column in self.feature_table.columns[2:]:
            # if 'Peak m/z' in column and column.split(' ')[-1] == 'm/z':
            #     column_mapping[column.split('.')[0]]['mz'] = column
            # elif 'Peak RT' in column and column.split(' ')[-1] == 'RT':
            #     column_mapping[column.split('.')[0]]['RT'] = column
            if 'Peak rtmin' in column:
                column_mapping[column.split('.')[0]]['rt_start'] = column
            elif 'Peak rtmax' in column:
                column_mapping[column.split('.')[0]]['rt_end'] = column
            elif 'Peak mzmin' in column:
                column_mapping[column.split('.')[0]]['mz_min'] = column
            elif 'Peak mzmax' in column:
                column_mapping[column.split('.')[0]]['mz_max'] = column
            elif 'Peak area' in column:
                column_mapping[column.split('.')[0]]['area'] = column
        self.column_map = column_mapping
        return column_mapping


    def load_features(self, sample_list):
        feature_number = self.feature_table.shape[0]
        i = 0
        # Iterate through all features in the feature table
        for index, row in self.feature_table.iterrows():
            i += 1
            # Extract feature specific information (mz and RT, RT is converted in seconds)
            mz = row['mz']
            RT = row['rt'] 
            feature_collection = FeatureCollection()
            # Iterate through all samples in the experiment
            for sample in sample_list:
                # Extract peak specific data for individual sample
                peak_dict = dict(
                    sample = sample,
                    RT = RT, 
                    mz = mz,
                    rt_start = row[self.column_map[sample.name]['rt_start']],
                    rt_end = row[self.column_map[sample.name]['rt_end']],
                    mz_min = row[self.column_map[sample.name]['mz_min']],
                    mz_max = row[self.column_map[sample.name]['mz_max']],
                    height = None,
                    area = row[self.column_map[sample.name]['area']]
                )
                # We do not save the peak if it does not exist in this sample (i.e. rt_end == 0 in table)
                if peak_dict['rt_end'] != 0:
                    # Create feature (No secondary id)
                    feature = Feature(mz, RT, sample, feature_collection=feature_collection)
                    # Create peak
                    new_peak = Peak(sample, peak_dict['RT'], peak_dict['mz'], peak_dict['rt_start'], peak_dict['rt_end'], peak_dict['mz_min'], 
                        peak_dict['mz_max'], peak_dict['height'], peak_dict['area'])
                    # Add feature to the peak (one peak is attached to a unique feature)
                    new_peak.feature = feature
                    # Add the peak to the sample
                    sample.peak_list.append(new_peak)
                    # Add the peak to the feature
                    feature.peak_list.append(new_peak)
                    # Add the feature to the sample
                    sample.feature_list.append(feature)
                    # Add the feature to the feature collection (Only one feature per feature collection in this case)
                    feature_collection.feature_list.append(feature)
            # Add the feature to the feature list (all peaks belonging to this feature have now been created)
            self.feature_collection_list.append(feature_collection)
        return None


class XcmsFeatureTable(FeatureTable):

    """
    Inherits from FeatureTable.

    XCMS specific feature table object.
    Enable reading XCMS unaligned feature tables.
    Aligned features not yet supported for XCMS
    """

    def __init__(self, feature_table_path=None, origin=None):
        logger.info('Feature table format: XCMS')
        super().__init__(feature_table_path, origin)


    def load_feature_table(self):
        feature_table_file_folder = pathlib.Path(self.feature_table_path)
        if self.feature_table_path == None:
            logger.error('Feature table path missing')
            return None
        # If the path is a directory (Unaligned peaks, one table per sample)
        elif feature_table_file_folder.is_dir():
            logger.error('Only .csv file is supported as XCMS feature table, directory given')
            return None
        else:
            feature_table = pd.read_csv(self.feature_table_path)
            self.feature_table = feature_table
            return feature_table


    def create_column_map(self, samples):
        # XCMS input is long format, we only need to map the sample names
        return None


    def load_features(self, sample_list):
        sample_map = dict()
        for sample in sample_list:
            sample_map[sample.file_name] = sample

        peak_number = self.feature_table.shape[0]
        feature_number = self.feature_table.index.max()
        i = 0

        logger.info('Loading %d features and %d peaks ', feature_number, peak_number)
        # Iterate through all features in the feature table
        for index, row in self.feature_table.iterrows():
            i += 1
            # Extract feature specific information (mz and RT, RT is converted in seconds)
            mz = row['mz']
            RT = row['rt'] / 60
            # No alignment, one feature per feature collection
            feature_collection = FeatureCollection()
            # Create feature (No secondary id)
            sample = sample_map[row['sample_name']]
            feature = Feature(mz, RT, sample, feature_collection=feature_collection)
            peak_dict = dict(
                sample = sample,
                RT = RT, 
                mz = mz,
                rt_start = row['rtmin'] / 60,
                rt_end = row['rtmax'] / 60,
                mz_min = row['mzmin'],
                mz_max = row['mzmax'],
                height = row['maxo'],
                area = row['into'],
                area_bc = row['intb'],
                sn = row['sn']
            )      
            # Create peak
            new_peak = Peak(sample, peak_dict['RT'], peak_dict['mz'], peak_dict['rt_start'], peak_dict['rt_end'], peak_dict['mz_min'], 
                peak_dict['mz_max'], peak_dict['height'], peak_dict['area'], area_baseline_corrected=peak_dict['area_bc'], sn=peak_dict['sn'])
            # Add feature to the peak (one peak is attached to a unique feature)
            new_peak.feature = feature
            # Add the peak to the sample
            sample.peak_list.append(new_peak)
            # Add the peak to the feature
            feature.peak_list.append(new_peak)
            # Add the feature to the sample
            sample.feature_list.append(feature)
            # Add the feature to the feature collection (Only one feature per feature collection in this case)
            feature_collection.feature_list.append(feature)
            # Add the feature to the feature list (all peaks belonging to this feature have now been created)
            self.feature_collection_list.append(feature_collection)
        logger.info('Feature table loaded with success')
        return None



import numpy as np

class Chromatogram():

    """
    Representation of a chromatographic peak

    Contains the related peak object, the size and margin of the chromatogram and its dataframe as a numpy array.
    """

    def __init__(self, peak, size, margin, dataframe):
        self.peak = peak
        self.size = size
        self.margin = margin
        self.dataframe = dataframe



class Peak():

    """
    Representation of a chromatographic peak.

    Contains the related feature and sample it belongs to, as well as all peak characteristics.
    """

    # Peak counter to automatically assign peak ids
    peak_counter = 0
    peak_dict = dict()

    # This class method returns the peak instance from a peak id
    # This is to avoid iteration through all instances of peaks to find a specific one
    @classmethod
    def get_peak(cls, id):
        return cls.peak_dict[id]


    def __init__(self, sample, RT, mz, rt_start, rt_end, mz_min, mz_max, height, area, feature = None, monoisotopic= None, valid=None, area_baseline_corrected=None, sn=None):
        self.id = Peak.peak_counter
        self.sample = sample
        self.RT = RT
        self.mz = mz
        self.rt_start = rt_start
        self.rt_end = rt_end
        self.mz_min = mz_min
        self.mz_max = mz_max
        self.height = height
        self.area = area
        self.feature = feature
        self.annotation = None
        self.scan_number = None
        self.monoisotopic = monoisotopic
        self.valid = valid
        self.area_baseline_corrected = area_baseline_corrected
        self.sn = sn
        self.formatted_chromatograms = []
        Peak.peak_dict[self.id] = self
        Peak.peak_counter += 1            


    def get_height(self):
        if self.height == None:
            # Deprecated, will be removed from next version on
            self.height = max(self.get_chromatogram_from_openms()['I'])
        return self.height


    def get_window_margin(self, margin=None):
        # Need to define a margin window around this to see what is around the peak
        if (margin != None) and (margin != 0):
            window_size = self.rt_end - self.rt_start
            rt_start = self.rt_start - (window_size*margin)
            rt_end = self.rt_end + (window_size*margin)
        else:
            rt_start = self.rt_start
            rt_end = self.rt_end
        return rt_start, rt_end


    def get_scan_number(self):
        if scan_number == None:
            chromatogram = self.sample.MS1[(self.sample.MS1['mz'] >= self.mz_min) & 
                              (self.sample.MS1['mz'] <= self.mz_max) & 
                              (self.sample.MS1['rt'] >= self.rt_start) & 
                              (self.sample.MS1['rt'] <= self.rt_end)]
            # Return the number of unique RT values which correspond to the number of scans 
            return chromatogram.rt.nunique()
        else:
            return self.scan_number


    def set_scan_number(self):
        if self.scan_number == None:
            chromatogram = self.sample.MS1[(self.sample.MS1['mz'] >= self.mz_min) & 
                              (self.sample.MS1['mz'] <= self.mz_max) & 
                              (self.sample.MS1['rt'] >= self.rt_start) & 
                              (self.sample.MS1['rt'] <= self.rt_end)]
            # Set scan number as the number of unique RT values which correspond to the number of scans 
            self.scan_number = chromatogram.rt.nunique()


    def get_chromatogram(self, margin=None):
        rt_start, rt_end = self.get_window_margin(margin)
        # We extract the values inside the peak window (mz and RT window)
        chromatogram = self.sample.extract_chromatogram(rt_start, rt_end, self.mz_min, self.mz_max)
        # Containt [rt, intensity, mz]
        return chromatogram


    def interpolate_chromatogram(self, vals, margin=1, normalise=True):
        # Get the full chromatogram with margin
        chromatogram = self.get_chromatogram(margin)
        # Remove the baseline
        intensity_array = chromatogram[1] - np.amin(chromatogram[1])
        # Get peak retention time start and end (including margin)
        rt_start, rt_end = self.get_window_margin(margin)
        # Prepare our evenly spaced numbers over the retention time interval for interpoaltion
        xvals = np.linspace(rt_start, rt_end, vals)
        # Interpolate intensity values, anything outside is set to 0
        yinterp = np.interp(xvals, chromatogram[0], intensity_array, left=0, right=0)
        # We scale the interpolated intensity values between 0 and 1 (Important)
        if normalise:
            yinterp = np.divide(yinterp, yinterp.max())
        # Create our array representing the peak window (1 = within the rt window, 0 = outside the rt window)
        window = [1 if (rt >= self.rt_start) and (rt <= self.rt_end) else 0 for rt in xvals]
        # Create our final matrix
        interpolated_chromatogram = np.array([yinterp,window])
        # Create a chromatogram object to store this matrix
        chromatogram_object = Chromatogram(self, vals, margin, interpolated_chromatogram)
        # Add it to the list of chromatograms available for this peak (for dev only to test different margin and matrix sizes)
        self.formatted_chromatograms.append(chromatogram_object)
        return interpolated_chromatogram


    def get_formatted_chromatogram(self, vals, margin):
        for chromatogram in self.formatted_chromatograms:
            if (chromatogram.size == vals) and (chromatogram.margin == margin):
                return chromatogram.dataframe
        chromatogram = self.interpolate_chromatogram(vals, margin, True)
        return chromatogram


    def formatted_chromatogram_exists(self, vals, margin):
        for chromatogram in self.formatted_chromatograms:
            if (chromatogram.size == vals) and (chromatogram.margin == margin):
                return True
        return False

